package fr.inria.phoenix.diasuite.generator.template.context

import fr.inria.phoenix.diasuite.diaswarm.BehaviorPublication
import fr.inria.phoenix.diasuite.diaswarm.ContextDef
import fr.inria.phoenix.diasuite.diaswarm.DataRequirement
import fr.inria.phoenix.diasuite.diaswarm.DataTypeRef
import fr.inria.phoenix.diasuite.diaswarm.DeviceType
import fr.inria.phoenix.diasuite.diaswarm.InteractionContractDef
import fr.inria.phoenix.diasuite.diaswarm.PullInteractionContract
import fr.inria.phoenix.diasuite.diaswarm.PushInteractionContract
import fr.inria.phoenix.diasuite.diaswarm.SourceDef
import fr.inria.phoenix.diasuite.diaswarm.SourceRef
import fr.inria.phoenix.diasuite.generator.template.IConstant
import java.util.ArrayList
import java.util.HashSet
import java.util.List
import java.util.Set
import org.eclipse.emf.ecore.EObject

import static extension fr.inria.phoenix.diasuite.generator.template.datatype.DiaSwarmDatatypeHelper.*

class DiaSwarmContextHelper {

	/**============================ HELPER METHODS ==============================**/
	def static generateFileName(ContextDef context) {
		context.getPackageName.replaceAll("\\.", "/") + "/" + IConstant.ABSTRACT_CLASS_PREFIX + context.name +
			IConstant.JAVA_EXTENSION_POSTFIX
	}

	def static getPackageName(ContextDef context) {
		IConstant.BASE_PACKAGE_NAMESPACE + "." + IConstant.CONTEXT_PACKAGE_NAME + "." + context.getName().toLowerCase()
	}

//	def static getPackageNameForContextFeature(ContextDef context, String contextFeature) {
//		IConstant.BASE_PACKAGE_NAMESPACE + "." + IConstant.CONTEXT_PACKAGE_NAME + "." +
//			context.getName().toLowerCase() + "." + contextFeature
//	}
	def static generateFileNameForContextFeature(ContextDef context, String contextFeature) {
		context.getPackageName.replaceAll("\\.", "/") + "/" + context.name.toFirstUpper + contextFeature +
			IConstant.JAVA_EXTENSION_POSTFIX
	}

	/**
	 * Return the full list of sources (including those defined in interaction contracts) used by a context
	 * 
	 * @param context
	 *            The context to extract sources from
	 * @return The full list of sources used by the context
	 */
	def static Set<SourceRef> getAllSources(ContextDef context) {
		val Set<SourceRef> processedSources = new HashSet<SourceRef>();

		// If the interaction is periodic or provided then we try to recover 
		// the pushed source if there is one.		
		for (interactionContract : context.behaviors)
			if (!interactionContract.interaction.equals(IConstant.REQUIRED_INTERACTION)) {
				var pushIC = (interactionContract.contract as PushInteractionContract)
				var source = pushIC.providedData.dataSource.sourceRef
				addSourceToSet(source, processedSources);

			} else {
				// Else, we know that all the sources if any are available through 
				// the required interaction.
				if (interactionContract.contract != null) {
					var pullIC = (interactionContract.contract as PullInteractionContract)

					for (requirement : pullIC.dataRequirements) {
						var source = requirement.dataSource.sourceRef
						addSourceToSet(source, processedSources);
					}
				}
			}

		processedSources;
	}

	def static addSourceToSet(SourceRef source, Set<SourceRef> processedSources) {
		if (source != null)
			if (!processedSources.contains(source))
				processedSources.add(source)
	}

	def static List<SourceRef> providedDeviceSources(ContextDef context) {
		val List<SourceRef> providedSources = new ArrayList<SourceRef>;

		for (InteractionContractDef interactionContract : context.behaviors) {
			val providedSource = (interactionContract.contract as PushInteractionContract).providedData.dataSource.
				sourceRef
			if (providedSource != null)
				providedSources.add(providedSource)
		}

		providedSources;
//		
//		.contract as PushInteractionContract)
//		interactionContract.providedData.dataSource.sourceRef;
	}

	def static Iterable<ContextDef> providedContexts(ContextDef context) {
		context.behaviors.filter [
			(!interaction.equals(IConstant.REQUIRED_INTERACTION)) &&
				((contract as PushInteractionContract).providedData.dataSource.context != null)
		].map[(contract as PushInteractionContract).providedData.dataSource.context];

//		val interactionContract = (context.behavior.contract as PushInteractionContract)
//		interactionContract.providedData.dataSource.context
	}

	def static String getContractReturnType(ContextDef context, BehaviorPublication publication, boolean inGet) {
		return switch (publication) {
			case BehaviorPublication::ALWAYS:
				if (context.indices.nullOrEmpty || inGet)
					context.dataTypeRef.fullName
				else
					context.name + "Value"
			case BehaviorPublication::MAYBE:
				context.name + "Value"
			default:
				if(inGet) context.dataTypeRef.fullName else "void"
		};
	}

	def static serviceDiscoveryRequired(String interaction, EObject contract) {
		// when required ...
		if (interaction.equals(IConstant.REQUIRED_INTERACTION)) {
			// if when required is followed by get ...
			(contract as PullInteractionContract) != null
		} // when provided || when periodic
		else
			(contract as PushInteractionContract).pullInteractionContract != null
	}

	def static getArgumentsForRequiredInteraction(String indices, Boolean discoveryRequired) {
		var arguments = ""

		if (!indices.empty) {
			arguments = arguments.concat(indices)
			if (discoveryRequired)
				arguments = arguments.concat(",\n        Discover discover")
		} else {
			if(discoveryRequired) arguments = arguments.concat("Discover discover")
		}

		arguments
	}

	def static String sourceFromDevice(
		SourceRef sourceRef) '''«sourceRef.source.name.toFirstUpper»From«sourceRef.device.name»'''

	/**
	 * Returns the device that the context needs to subscribe to.
	 * This situation applies to when provided/periodic interaction contracts.
	 */
	def static DeviceType deviceForSubscription(InteractionContractDef interactionContract) {
		val contract = (interactionContract.contract as PushInteractionContract)
		val sourceReference = contract.providedData.dataSource.sourceRef

		return if (sourceReference != null)
			sourceReference.device
		else
			null
	}

	def static Iterable<SourceDef> providedSources(ContextDef context) {
		return context.behaviors.filter [
			(!interaction.equals(IConstant.REQUIRED_INTERACTION)) &&
				((contract as PushInteractionContract).providedData.dataSource.sourceRef != null)
		].map[(contract as PushInteractionContract).providedData.dataSource.sourceRef.source];

//		val contract = (context.behavior.contract as PushInteractionContract)
//		val sourceReference = contract.providedData.dataSource.sourceRef
//
//		return if (sourceReference != null)
//			sourceReference.source
//		else
//			null
	}

	def static getDataRequirementsForGet(ContextDef context) {
		var List<DataRequirement> dataRequirements = new ArrayList<DataRequirement>;

		for (InteractionContractDef behavior : context.behaviors) {
			if ((behavior.interaction.equals(IConstant.REQUIRED_INTERACTION))) {
				if ((behavior.contract != null))
					dataRequirements.addAll((behavior.contract as PullInteractionContract).dataRequirements)
			} else {
				val pullInteractionContract = (behavior.contract as PushInteractionContract).pullInteractionContract
				if (pullInteractionContract != null)
					dataRequirements.addAll(pullInteractionContract.dataRequirements)
			}
		}

		dataRequirements
	}

	def static getInteractionContractsForGet(ContextDef context) {
		var List<InteractionContractDef> interactionContracts = new ArrayList<InteractionContractDef>;

		for (InteractionContractDef behavior : context.behaviors) {
			if ((behavior.interaction.equals(IConstant.REQUIRED_INTERACTION))) {
				if ((behavior.contract != null))
					interactionContracts.addAll(behavior)
			} else {
				val pullInteractionContract = (behavior.contract as PushInteractionContract).pullInteractionContract
				if (pullInteractionContract != null)
					interactionContracts.addAll(behavior)
			}
		}

		interactionContracts
	}

	def static hasGroupingAttribute(PushInteractionContract contract) {
		return ( contract.providedData.groupingAttribute != null)
	}

	def static isCustomType(DataTypeRef dataTypeRef) {
		return dataTypeRef.type != null
	}

	def static getDeliveryModelsMap(String interaction, String providedSourceName, int period, String timeUnit, int valuesNumber) {
		if (interaction.equals(IConstant.PERIODIC_INTERACTION))
			return '''		
				deliveryModels.put("«providedSourceName»",
					new DataDeliveryModel.Builder()
						.deliveryModel(DataDeliveryEnum.PERIODIC)
							.period(«period»).timeUnit(TimeUnit.«timeUnit»).valueCount(«valuesNumber»).build());
			'''
		else
			return '''
				deliveryModels.put("«providedSourceName»",
							new DataDeliveryModel.Builder()
									.deliveryModel(DataDeliveryEnum.EVENT_DRIVEN)
									.build());
			'''
	}

	def static getContextDataSourceName(PushInteractionContract interactionContract) {
		val sourceRef = interactionContract.providedData.dataSource.sourceRef
		val contextRef = interactionContract.providedData.dataSource.context
		var String dataSourceName = null

		if (sourceRef != null)
			dataSourceName = sourceRef.source.name
		else
			dataSourceName = contextRef.name

		return dataSourceName
	}

	def static mapReduce(ContextDef context) {
		for (InteractionContractDef interactionContract : context.behaviors) {
			if (interactionContract.interaction.equals(IConstant.PERIODIC_INTERACTION)) {
				val map = (interactionContract.contract as PushInteractionContract).providedData.map
				val reduce = (interactionContract.contract as PushInteractionContract).providedData.reduce

				return if (map && reduce) '''implements MapReduce<«interactionContract.mapReduceTypes»>''' else ''''''
			}
		}

		return ''''''
	}

	def static mapReduceTypes(InteractionContractDef interactionContract) {
		val pushIC = (interactionContract.contract as PushInteractionContract)
		val groupingAttribute = pushIC.providedData.groupingAttribute;
		var String mapKeyInputType = if(groupingAttribute != null) groupingAttribute.attrTypeRef.fullName else "String";
		var String mapValueInputType = pushIC.providedData.dataSource.sourceRef.source.dataTypeRef.fullName
		var String mapKeyOutputType = mapKeyInputType
//		var String mapValueOutputType = pushIC.providedData.mapType.literal
		var String mapValueOutputType = pushIC.providedData.mapType.typeName
		var String reduceKeyOutPutType = mapKeyOutputType
		var String reduceValueOutPutType = pushIC.providedData.reduceType.typeName
//		var String reduceValueOutPutType = pushIC.providedData.reduceType.literal

		return mapKeyInputType.split('\\.').last + ", " + mapValueInputType.split('\\.').last + ", " +
			mapKeyOutputType.split('\\.').last + ", " + mapValueOutputType.split('\\.').last + ", " +
			reduceKeyOutPutType.split('\\.').last + ", " + reduceValueOutPutType.split('\\.').last
	}

	def static mapReduceTypes(PushInteractionContract interactionContract) {
		val pushIC = (interactionContract)
		val groupingAttribute = pushIC.providedData.groupingAttribute;
		var String mapKeyInputType = if(groupingAttribute != null) groupingAttribute.attrTypeRef.fullName else "String";
		var String mapValueInputType = pushIC.providedData.dataSource.sourceRef.source.dataTypeRef.fullName
		var String mapKeyOutputType = mapKeyInputType
		var String mapValueOutputType = pushIC.providedData.mapType.typeName
//		var String mapValueOutputType = pushIC.providedData.mapType.literal
		var String reduceKeyOutPutType = mapKeyOutputType
		var String reduceValueOutPutType = pushIC.providedData.reduceType.typeName
//		var String reduceValueOutPutType = pushIC.providedData.reduceType.literal

		return mapKeyInputType.split('\\.').last + ", " + mapValueInputType.split('\\.').last + ", " +
			mapKeyOutputType.split('\\.').last + ", " + mapValueOutputType.split('\\.').last + ", " +
			reduceKeyOutPutType.split('\\.').last + ", " + reduceValueOutPutType.split('\\.').last
	}

	def static mapReduceBehaviors(ContextDef context) {
		return context.behaviors.filter [
			(interaction.equals(IConstant.PERIODIC_INTERACTION)) &&
				(contract as PushInteractionContract).providedData.map &&
				(contract as PushInteractionContract).providedData.reduce
		]
	}

//	def static Set<DeviceType> discoverableDevices(ContextDef context) {
//		val interaction = context.behavior.interaction
//		val contract = context.behavior.contract
//		val Set<DeviceType> discoverableDevices = new HashSet<DeviceType>;
//
//		if (interaction.equals(IConstant.REQUIRED_INTERACTION)) {
//			if (contract != null) {
//				val dataRequirements = (contract as PullInteractionContract).dataRequirements
//				discoverableDevices.addAll(dataRequirements.devicesFromDataRequirements)
//			}
//
//		} else if (interaction.equals(IConstant.PROVIDED_INTERACTION) ||
//			interaction.equals(IConstant.PERIODIC_INTERACTION)) {
//			val pullInteractionContract = (contract as PushInteractionContract).pullInteractionContract
//
//			if (pullInteractionContract != null) {
//				val dataRequirements = pullInteractionContract.dataRequirements
//				discoverableDevices.addAll(dataRequirements.devicesFromDataRequirements)
//			}
//		}
//
//		discoverableDevices
//	}
//	def static getDevicesFromDataRequirements(EList<DataRequirement> dataRequirements) {
//		val Set<DeviceType> discoverableDevices = new HashSet<DeviceType>;
//
//		for (requirement : dataRequirements) {
//			val dataSource = requirement.dataSource
//			if (dataSource.sourceRef != null)
//				discoverableDevices.add(dataSource.sourceRef.device)
//		}
//
//		discoverableDevices
//	}
//	def static SourceDef discoverableSources(ContextDef context, DeviceType device) {
//		return providedSource(context).filter([device == device]).map([source]).toSet
//	}
/*============================================================================*/
}